# Standard lib imports
import random
import re


# Local imports
from qwilfish.constants import DEFAULT_START_SYMBOL
from qwilfish.grammar import opts

# TODO fuzz by changing endianness of fields and/or bytes
# TODO fuzz by inserting wrong TLVs

HEX_MAP = {"0": "0000", "1": "0001", "2": "0010", "3": "0011",
           "4": "0100", "5": "0101", "6": "0110", "7": "0111",
           "8": "1000", "9": "1001", "A": "1010", "B": "1011",
           "C": "1100", "D": "1101", "E": "1110", "F": "1111"}

def srange(characters):
    return [c for c in characters]

def to_binstr(ethernet_frame, subtree=None):
    symbol, children = subtree

    if children == []:
        if symbol == "x":
            return ("", []) # Turn single x:es into null expansions to remove
        else:
            new_symbol = \
                re.sub("x[0-9a-fA-F]{2}",
                       lambda hex: HEX_MAP[hex.group(0)[1:].upper()[0]] +
                                   HEX_MAP[hex.group(0)[1:].upper()[1]],
                       symbol)
            return (new_symbol, [])
    else:
        if symbol == "<hex>":
            child_symbol, _ = children[0]
            return (symbol, [(HEX_MAP[child_symbol], [])])

    return (symbol, [to_binstr("", c) for c in children if children])

def gen_lv():
    length = random.randint(16, 16) # TODO probability distribution?
    bin_length = ["1" if length & (1 << i) else "0" for i in range(0, 9)]
    bin_value = [repr(random.randint(0,1)) for b in range(0, length*8)]
    return "".join(bin_length) + "".join(bin_value)

ETHERNET_FRAME_GRAMMAR = {
    DEFAULT_START_SYMBOL  : [("<ethernet-frame>", opts(post=to_binstr))],
    "<ethernet-frame>"    : ["<addr><vlan-tags><type-payload>"],
    "<addr>"              : ["<dst><src>"],
    "<dst>"               : ["<byte><byte><byte><byte><byte><byte>",
                             "<mef-multicast>"
                            ],
    "<mef-multicast>"     : ["x01x80xC2x00x00x00", "x01x80xC2x00x00x03",
                             "x01x80xC2x00x00x0E"
                            ],
    "<src>"               : ["<byte><byte><byte><byte><byte><byte>"],
    "<vlan-tags>"         : ["", "<q-tag><vlan-tags>", "<q-tag>"],
    "<q-tag>"             : ["<tpid><pcp><dei><vlan>"],
    "<tpid>"              : ["x81x00", "x88xA8"],
    "<pcp>"               : ["<bit><bit><bit>"],
    "<dei>"               : ["<bit>"],
    "<vlan>"              : ["<byte><bit><bit><bit><bit>"],
    "<type-payload>"      : ["<lldp-ethertype><lldp-payload>"],
    "<byte>"              : ["x<hex><hex>"],
    "<hex>"               : srange("0123456789ABCDEF"),
    "<bit>"               : ["0", "1"],
    "<lldp-ethertype>"    : ["x88xCC"],
    "<lldp-payload>"      : ["<lldp-chassiid-tlv><lldp-portid-tlv>" \
                             "<lldp-ttl-tlv><lldp-opt-tlvs><lldp-end-tlv>"],
    "<lldp-end-tlv>"      : ["x00x00"],
    "<lldp-chassiid-tlv>" : ["0000001<lldp-lv>"],
    "<lldp-portid-tlv>"   : ["0000010<lldp-lv>"],
    "<lldp-ttl-tlv>"      : ["0000011<lldp-lv>"],
    "<lldp-opt-tlvs>"     : ["", "<lldp-opt-tlv>",
                             "<lldp-opt-tlv><lldp-opt-tlvs>"
                            ],
    "<lldp-opt-tlv>"      : ["<lldp-portdesc-tlv>", "<lldp-sysname-tlv>",
                             "<lldp-sysdesc-tlv>", "<lldp-syscap-tlv>",
                             "<lldp-mgmtaddr-tlv>", "<lldp-res-tlv>",
                             "<lldp-custom-tlv>"
                            ],
    "<lldp-portdesc-tlv>" : ["0000100<lldp-lv>"],
    "<lldp-sysname-tlv>"  : ["0000101<lldp-lv>"],
    "<lldp-sysdesc-tlv>"  : ["0000110<lldp-lv>"],
    "<lldp-syscap-tlv>"   : ["0000111<lldp-lv>"],
    "<lldp-mgmtaddr-tlv>" : ["0001000<lldp-lv>"],
    "<lldp-res-tlv>"      : ["<bit><bit><bit><bit><bit><bit><bit><lldp-lv>"],
    "<lldp-custom-tlv>"   : ["1111111<lldp-lv>"],
    "<lldp-lv>"           : [("<byte>", opts(pre=gen_lv))],
}



'''
Constants used within the package
'''

DEFAULT_START_SYMBOL = "<start>"
DEFAULT_MIN_NONTERMINALS = 0
DEFAULT_MAX_NONTERMINALS = 10

REGEX_NONTERMINAL_STR = r"(<[^<> ]*>)"

SIMPLE_GRAMMAR_EXAMPLE = {
    DEFAULT_START_SYMBOL: ["<order>"],
    "<order>": ["<greeting><order-list><goodbye>"],
    "<greeting>": ["Hello, I would like ", "Give me "],
    "<order-list>": ["<dish> and <order-list>", "<dish>."],
    "<dish>": ["a hamburger", "the pasta", "a kebab", "a pizza",
               "today's special"],
    "<goodbye>": [" Thanks!", ""]
}



from qwilfish.qwilfuzzer import QwilFuzzer
from qwilfish.constants import SIMPLE_GRAMMAR_EXAMPLE
from qwilfish.ethernet_frame import ETHERNET_FRAME_GRAMMAR
from qwilfish.linux_socket_runner import LinuxSocketRunner

def main():
    grammar = ETHERNET_FRAME_GRAMMAR
    qf = QwilFuzzer(grammar, log=True)
    qf.run(runner=LinuxSocketRunner())



# Standard lib imports
import re
import sys

# Local imports
from qwilfish.constants import DEFAULT_START_SYMBOL

NONTERMINAL_REGEX_STR = r"(<[^<> ]*>)"
RE_NONTERMINAL = re.compile(NONTERMINAL_REGEX_STR)

EXP_STRING_INDEX = 0
EXP_OPTS_INDEX = 1

class GrammarError(Exception):
    '''Custom exception class for errors related to the usage of a grammar'''
    pass

def validate(grammar,
             start_symbol=DEFAULT_START_SYMBOL,
             supported_opts=None):

    if not isinstance(grammar, dict):
        return False

    if len(grammar) == 0:
        return False

    defined_nonterminals, used_nonterminals = \
        _defined_and_used_nonterminals(grammar, start_symbol)

    if defined_nonterminals is None or used_nonterminals is None:
        return False

    # Always consider the "<start>" symbol as being used
    if DEFAULT_START_SYMBOL in grammar:
        used_nonterminals.add(DEFAULT_START_SYMBOL)

    # Find undefined and unused symbols
    for unused_nonterminal in defined_nonterminals - used_nonterminals:
        print(repr(unused_nonterminal) + ": defined, but not used",
              file=sys.stderr)
    for undefined_nonterminal in used_nonterminals - defined_nonterminals:
        print(repr(undefined_nonterminal) + ": used, but not defined",
              file=sys.stderr)

    # Find unreachable symbols
    unreachable = _unreachable_nonterminals(grammar, start_symbol)
    msg_start_symbol = start_symbol
    if DEFAULT_START_SYMBOL in grammar:
        unreachable = unreachable - \
            _reachable_nonterminals(grammar, DEFAULT_START_SYMBOL)
        if start_symbol != DEFAULT_START_SYMBOL:
            msg_start_symbol += " or " + DEFAULT_START_SYMBOL
    for unreachable_nonterminal in unreachable:
        print(repr(unreachable_nonterminal) + \
              ": unreachable from " + msg_start_symbol,
              file=sys.stderr)

    # Find out what options were used and if they were supported
    used_but_not_supported_opts = set()
    if supported_opts is not None:
        used_but_not_supported_opts = opts_used(
            grammar, start_symbol).difference(supported_opts)
        for opt in used_but_not_supported_opts:
            print("warning: option " + repr(opt) + " is not supported",
                  file=sys.stderr)

    return used_nonterminals == defined_nonterminals and \
           len(unreachable) == 0

def nonterminals(expansion):
    if isinstance(expansion, tuple):
        expansion = expansion[0]

    return re.findall(RE_NONTERMINAL, expansion)

def is_nonterminal(s):
    return re.match(RE_NONTERMINAL, s)

def opts(**kwargs):
    return kwargs

def opts_used(grammar, start_symbol):
    used_opts = set()
    for symbol in grammar:
        for expansion in grammar[symbol]:
            used_opts |= set(expansion_opts(expansion).keys())
    return used_opts

def expansion_string(expansion):
    return _expansion_element(expansion, EXP_STRING_INDEX)

def expansion_opts(expansion):
    return _expansion_element(expansion, EXP_OPTS_INDEX)

def expansion_opt(expansion, attribute):
    return expansion_opts(expansion).get(attribute, None)

# Private/helper methods below

def _defined_and_used_nonterminals(grammar, start_symbol):
    defined_nonterminals = set()
    used_nonterminals = {start_symbol}

    for defined_nonterminal in grammar:
        defined_nonterminals.add(defined_nonterminal)
        expansions = grammar[defined_nonterminal]
        if not isinstance(expansions, list):
            print(repr(defined_nonterminal) + ": expansion is not a list",
                  file=sys.stderr)
            return None, None

        if len(expansions) == 0:
            print(repr(defined_nonterminal) + ": expansion list is empty",
                  file=sys.stderr)
            return None, None

        for expansion in expansions:
            if isinstance(expansion, tuple):
                expansion = expansion[0]
            if not isinstance(expansion, str):
                print(repr(defined_nonterminal) + ": " +
                      repr(expansion) + "not a string",
                      file=sys.stderr)
                return None, None

            for used_nonterminal in nonterminals(expansion):
                used_nonterminals.add(used_nonterminal)

    return defined_nonterminals, used_nonterminals

def _unreachable_nonterminals(grammar, start_symbol):
    return grammar.keys() - _reachable_nonterminals(grammar, start_symbol)

def _reachable_nonterminals(grammar, start_symbol):
    reachable = set()

    def _reachable_nonterminals_inner(grammar, symbol):
        nonlocal reachable
        reachable.add(symbol)
        for expansion in grammar.get(symbol, []):
            for nonterminal in nonterminals(expansion):
                if nonterminal not in reachable:
                    _reachable_nonterminals_inner(grammar, nonterminal)

    _reachable_nonterminals_inner(grammar, start_symbol)

    return reachable

def _expansion_element(expansion, index=EXP_STRING_INDEX):
    if isinstance(expansion, str):
        return expansion if index == EXP_STRING_INDEX else {}
    elif isinstance(expansion, tuple) and \
         len(expansion) == 2 and \
         isinstance(expansion[EXP_STRING_INDEX], str):
        return expansion[index]
    else:
        raise TypeError(repr(expansion) + \
            "Expansion is not a 2-tuple with a str in position 0")



'''
Utilities for dealing with derivation trees represented as a tuple of
(node, children).
'''



# Standard library imports
import re

# Local imports
import qwilfish.grammar as qg

class DerivationTreeError(Exception):
    ''' Custom exception class for errors related to the derivation tree'''

def all_terminals(tree):
    '''
    Return the string tree represented by the tree thus far. Unexpanded
    nonterminals will be written as <nonterminal> in the resulting string.
    '''

    (symbol, children) = tree

    if children is None or len(children) == 0:
        return symbol

    return "".join([all_terminals(c) for c in children])

def expansion_to_children(expansion):
    '''
    Expand all symbols in an expansion string. Will return a list of
    2-tuples, one for each symbol in the expansion string. Nonterminals
    will be a tuple of the form ("<nonterminal>", None) and terminals will be
    of the form ("terminal", []).
    '''

    expansion = qg.expansion_string(expansion)

    if expansion == "": # Epsilon expansion
        return [("", [])]

    strings = re.split(qg.RE_NONTERMINAL, expansion)
    return [(s, None) if qg.is_nonterminal(s) else (s,[])
            for s in strings if len(s) > 0]
